@@ -1,6 +1,8 @@
 source common.sh
 
-set -o pipefail
+if [[ $BUILD_SHARED_LIBS != 1 ]]; then
+    skipTest "Plugins are not supported"
+fi
 
 res=$(nix --option setting-set true --option plugin-files $PWD/plugins/libplugintest* eval --expr builtins.anotherNull)
 
